#include "AgentStates.h"
#include "Agent.h"
#include <sstream>
#include "EntityManager.h"
#include "Messages.h"

AgentState_Idle* AgentState_Idle::Instance()
{
	static AgentState_Idle instance;
	return &instance;
}

void AgentState_Idle::Enter(Agent* owner)
{
	//Stop steering forces (except avoidance),
	owner->GetSteering()->ArriveOff();
	owner->GetSteering()->FleeOff();
	owner->GetSteering()->WanderOff();
	owner->GetSteering()->SeekOff();
}

void AgentState_Idle::Execute(Agent* owner, float dt)
{
	
}

void AgentState_Idle::Exit(Agent* owner)
{
	//Do nothing.

}

void AgentState_Idle::Render(Agent* owner)
{
	
}

AgentState_Move* AgentState_Move::Instance()
{
	static AgentState_Move instance;
	return &instance;
}

void AgentState_Move::Enter(Agent* owner)
{
	owner->GetSteering()->Arrive(owner->GetTargetPosition(), 10.0, 50.0f, 0.5f, 1.0f);
	//Stop steering forces (except avoidance),
	owner->GetSteering()->FleeOff();
	owner->GetSteering()->WanderOff();
	owner->GetSteering()->SeekOff();

}

void AgentState_Move::Execute(Agent* owner, float dt)
{
	owner->GetSteering()->Arrive(owner->GetTargetPosition(), 10.0, 50.0f, 0.5f, 1.0f);
	owner->GetPhysics()->Accelerate(owner->GetSteering()->Calculate().linear);
}

void AgentState_Move::Exit(Agent* owner)
{


}

void AgentState_Move::Render(Agent* owner)
{


}

AgentState_Retreat* AgentState_Retreat::Instance()
{
	static AgentState_Retreat instance;
	return &instance;
}

void AgentState_Retreat::Enter(Agent* owner)
{
	//Stop steering forces (except avoidance),
	owner->GetSteering()->Flee(owner->GetTargetPosition(), 1.0f);
	owner->GetSteering()->WanderOff();
	owner->GetSteering()->SeekOff();
	owner->GetSteering()->ArriveOff();
}

void AgentState_Retreat::Execute(Agent* owner, float dt)
{
	owner->GetSteering()->Flee(owner->GetTargetPosition(), 1.0f);
	owner->GetPhysics()->Accelerate(owner->GetSteering()->Calculate().linear);
}

void AgentState_Retreat::Exit(Agent* owner)
{


}

void AgentState_Retreat::Render(Agent* owner)
{


}

AgentState_Attack* AgentState_Attack::Instance()
{
	static AgentState_Attack instance;
	return &instance;
}

void AgentState_Attack::Enter(Agent* owner)
{
	owner->GetSteering()->Arrive(owner->GetTargetPosition(), 50.0, 100.0f, 0.5f, 1.0f);
	//Stop steering forces (except avoidance),
	owner->GetSteering()->FleeOff();
	owner->GetSteering()->WanderOff();
	owner->GetSteering()->SeekOff();

}

void AgentState_Attack::Execute(Agent* owner, float dt)
{
	//Update target
	owner->GetSteering()->Arrive(owner->GetTargetPosition(), 50.0, 100.0f, 0.5f, 1.0f);
	//Move at half speed.
	owner->GetPhysics()->Accelerate(0.5 * owner->GetSteering()->Calculate().linear);
	//Attack if ammo is available.
	if(owner->GetWeapon()->GetAmmoCount() > 0)
		owner->UseWeapon();
}

void AgentState_Attack::Exit(Agent* owner)
{


}

void AgentState_Attack::Render(Agent* owner)
{


}

AgentState_GlobalRed* AgentState_GlobalRed::Instance()
{
	static AgentState_GlobalRed instance;
	return &instance;
}

void AgentState_GlobalRed::Enter(Agent* owner)
{
	//Stop steering forces (except avoidance),
	owner->GetSteering()->FleeOff();
	owner->GetSteering()->WanderOff();
	owner->GetSteering()->SeekOff();
	//Load texture
	D3DXIMAGE_INFO m_ImageInfo;
	HR(D3DXCreateTextureFromFileEx(gD3DDevice, "Basic_Vehicle_Red.png", 0, 0, 0, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT, D3DCOLOR_ARGB(255,0,0,255), &m_ImageInfo, 0, &m_Texture));
}

void AgentState_GlobalRed::Execute(Agent* owner, float dt)
{
	if(owner->GetHealth() <= 0.0)
	{
		gMessenger->AddMessage(owner->GetID(), owner->GetID(), Message::msg_Destroy, 0.0f);
	}
}

void AgentState_GlobalRed::Exit(Agent* owner)
{


}

void AgentState_GlobalRed::Render(Agent* owner)
{
	std::stringstream ss;
	owner->GetSteering()->Render();
	D3DXMATRIX scaleMat, worldMat, transMat, rotMat;
	D3DXMatrixScaling(&scaleMat, 0.5f, 0.5f, 0.5f);
	D3DXMatrixTranslation(&transMat, owner->GetPhysics()->GetPosition().x, owner->GetPhysics()->GetPosition().y, 0.0f);
	D3DXMatrixRotationZ(&rotMat, owner->GetPhysics()->GetRotation());
	D3DXMatrixMultiply(&scaleMat, &scaleMat, &rotMat);
	D3DXMatrixMultiply(&worldMat, &scaleMat, &transMat);
	HR(gSprite->SetTransform(&worldMat));
	HR(gSprite->Draw(m_Texture, 0, &D3DXVECTOR3(32.0f, 32.0f, 0.0f), 0, D3DCOLOR_ARGB(255,255,255,255)));
	HR(gSprite->Flush());

	D3DXMatrixScaling(&scaleMat, 0.5f, 0.5f, 0.5f);
	D3DXMatrixTranslation(&transMat, owner->GetPhysics()->GetPosition().x - 16, owner->GetPhysics()->GetPosition().y - 40, 0.0f);
	D3DXMatrixMultiply(&worldMat, &scaleMat, &transMat);
	HR(gSprite->SetTransform(&worldMat));

	ss.precision(0);
	ss << "H:" << owner->GetHealth() << std::endl << "A:" << owner->GetWeapon()->GetAmmoCount();
	gApplication->GetDefaultFont()->DrawTextA(gSprite, ss.str().c_str(), -1, 0, DT_TOP | DT_LEFT , D3DCOLOR_ARGB(255,0,0,0));
}

AgentState_GlobalBlue* AgentState_GlobalBlue::Instance()
{
	static AgentState_GlobalBlue instance;
	return &instance;
}

void AgentState_GlobalBlue::Enter(Agent* owner)
{
	//Stop steering forces (except avoidance),
	owner->GetSteering()->FleeOff();
	owner->GetSteering()->WanderOff();
	owner->GetSteering()->SeekOff();
	//Load texture
	D3DXIMAGE_INFO m_ImageInfo;
	HR(D3DXCreateTextureFromFileEx(gD3DDevice, "Basic_Vehicle_Blue.png", 0, 0, 0, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT, D3DCOLOR_ARGB(255,0,0,255), &m_ImageInfo, 0, &m_Texture));
}

void AgentState_GlobalBlue::Execute(Agent* owner, float dt)
{
	if(owner->GetHealth() <= 0.0)
		gMessenger->AddMessage(owner->GetID(), owner->GetID(), Message::msg_Destroy, 0.0f);
}

void AgentState_GlobalBlue::Exit(Agent* owner)
{


}

void AgentState_GlobalBlue::Render(Agent* owner)
{
	std::stringstream ss;
	owner->GetSteering()->Render();
	D3DXMATRIX scaleMat, worldMat, transMat, rotMat;
	D3DXMatrixScaling(&scaleMat, 0.5f, 0.5f, 0.5f);
	D3DXMatrixTranslation(&transMat, owner->GetPhysics()->GetPosition().x, owner->GetPhysics()->GetPosition().y, 0.0f);
	D3DXMatrixRotationZ(&rotMat, owner->GetPhysics()->GetRotation());
	D3DXMatrixMultiply(&scaleMat, &scaleMat, &rotMat);
	D3DXMatrixMultiply(&worldMat, &scaleMat, &transMat);
	HR(gSprite->SetTransform(&worldMat));
	HR(gSprite->Draw(m_Texture, 0, &D3DXVECTOR3(32.0f, 32.0f, 0.0f), 0, D3DCOLOR_ARGB(255,255,255,255)));
	HR(gSprite->Flush());

	D3DXMatrixScaling(&scaleMat, 0.5f, 0.5f, 0.5f);
	D3DXMatrixTranslation(&transMat, owner->GetPhysics()->GetPosition().x - 16, owner->GetPhysics()->GetPosition().y - 40, 0.0f);
	D3DXMatrixMultiply(&worldMat, &scaleMat, &transMat);
	HR(gSprite->SetTransform(&worldMat));

	ss.precision(0);
	ss << "H:" << owner->GetHealth() << std::endl << "A:" << owner->GetWeapon()->GetAmmoCount();
	gApplication->GetDefaultFont()->DrawTextA(gSprite, ss.str().c_str(), -1, 0, DT_TOP | DT_LEFT , D3DCOLOR_ARGB(255,0,0,0));
}