#include "DirectX9App.h"
#include "DirectInput.h"
#include "DirectX9Game.h"
#include <ctime>


//Contains the main game loop. Initializes and shuts down all other systems.
int DirectX9App::Run()
{
	MSG  msg;
    msg.message = WM_NULL;

	__int64 cntsPerSec = 0;
	QueryPerformanceFrequency((LARGE_INTEGER*)&cntsPerSec);
	float secsPerCnt = 1.0f / (float)cntsPerSec;

	__int64 prevTimeStamp = 0;
	QueryPerformanceCounter((LARGE_INTEGER*)&prevTimeStamp);
		
	//Random number gen seed
	while(msg.message != WM_QUIT)
	{
		// If there are Window messages then process them.
		if(PeekMessage( &msg, 0, 0, 0, PM_REMOVE ))
		{
            TranslateMessage( &msg );
            DispatchMessage( &msg );
		}
		else
        {	
			// If the application is paused then free some CPU cycles to other 
			// applications and then continue on to the next frame.
			if( m_AppPaused )
			{
				Sleep(20);
				continue;
			}

			//Main Game Loop.
			if( !isDeviceLost() )
			{
				__int64 currTimeStamp = 0;
				QueryPerformanceCounter((LARGE_INTEGER*)&currTimeStamp);
				float dt = (currTimeStamp - prevTimeStamp)*secsPerCnt;
				//Update primary systems here
				gInput->Update();
				gMessenger->Update(dt);
				
				//Set Exit command
				if(gInput->KeyDown(DIK_ESCAPE))
				{
					PostQuitMessage(0);
				}
				//Update Game
				gGame->Update(dt);
				//Update window rect
				m_WindowRect.top = 0;
				m_WindowRect.left = 0;
				m_WindowRect.right = GetWindowSize().x;
				m_WindowRect.bottom = GetWindowSize().y;
				//Begin rendering
				BeginFrame();
				//Render Input related objects (cursor and/or mouse position)
				//gInput->Render();
				//Render Game
				gGame->Render();
				//End rendering
				EndFrame();
				prevTimeStamp = currTimeStamp;
			}
        }
    }
	//Shutdown primary systems
	gInput->Shutdown();
	return (int)msg.wParam;
}

//Release non mananged objects
void DirectX9App::onLostDevice()
{
	m_SpriteInterface->OnLostDevice();
}

//Reset non managed objects
void DirectX9App::onResetDevice()
{
	m_SpriteInterface->OnResetDevice();
}

//Link to App instance
DirectX9App* DirectX9App::Instance()
{
	static DirectX9App instance;
	return &instance;
}

DirectX9App::DirectX9App():
	m_MainWndCaption(APP_TITLE),
	m_hMainWnd(0),
	m_D3DObject(0),
	m_D3DDevice(0),
	m_ClearColor(1.0f,1.0f,1.0f,1.0f),		//Blue clear color
	m_AppPaused(false),
	m_SpriteInterface(0)
{
	//Data Member Initialization
	D3DXMatrixIdentity(&m_ProjMat);
	D3DXMatrixIdentity(&m_ViewMat);
	ZeroMemory(&m_D3DPP, sizeof(m_D3DPP));
	//Init Window
	InitMainWindow();
	//Init DirectX
	InitDirect3D();
	//Seed random numbers
	srand(time(NULL));
	//Load background
}

DirectX9App::~DirectX9App()
{
	Shutdown();
}

void DirectX9App::EnableFullScreenMode(bool enable)
{
	// Switch to fullscreen mode.
	if( enable )
	{
		// Are we already in fullscreen mode?
		if( !m_D3DPP.Windowed ) 
			return;

		//Get screen size
		int width  = GetSystemMetrics(SM_CXSCREEN);
		int height = GetSystemMetrics(SM_CYSCREEN);

		//Reset backbuffer
		m_D3DPP.BackBufferFormat = D3DFMT_X8R8G8B8;
		m_D3DPP.BackBufferWidth  = width;
		m_D3DPP.BackBufferHeight = height;
		m_D3DPP.Windowed         = false;

		// Change the window style to a more fullscreen friendly style.
		SetWindowLongPtr(m_hMainWnd, GWL_STYLE, WS_POPUP);
		SetWindowPos(m_hMainWnd, HWND_TOP, 0, 0, width, height, SWP_NOZORDER | SWP_SHOWWINDOW);	
	}
	// Switch to windowed mode.
	else
	{
		// Are we already in windowed mode?
		if( m_D3DPP.Windowed ) 
			return;

		//Reset Backbuffer
		RECT R = {0, 0, BASE_WINDOW_WIDTH, BASE_WINDOW_HEIGHT};
		AdjustWindowRect(&R, WS_OVERLAPPEDWINDOW, false);
		m_D3DPP.BackBufferFormat = D3DFMT_UNKNOWN;
		m_D3DPP.BackBufferWidth  = 1024;
		m_D3DPP.BackBufferHeight = 512;
		m_D3DPP.Windowed         = true;
	
		// Change the window style to a more windowed friendly style.
		SetWindowLongPtr(m_hMainWnd, GWL_STYLE, WS_OVERLAPPEDWINDOW);
		SetWindowPos(m_hMainWnd, HWND_TOP, 100, 100, R.right, R.bottom, SWP_NOZORDER | SWP_SHOWWINDOW);
	}


				

	// Reset the device with the changes.
	onLostDevice();
	HR(m_D3DDevice->Reset(&m_D3DPP));
	onResetDevice();
}

bool DirectX9App::isDeviceLost()
{
	// Get the state of the graphics device.
	HRESULT hr = gD3DDevice->TestCooperativeLevel();

	// If the device is lost and cannot be reset yet then
	// sleep for a bit and we'll try again on the next 
	// message loop cycle.
	if( hr == D3DERR_DEVICELOST )
	{
		Sleep(20);
		return true;
	}
	// Driver error, exit.
	else if( hr == D3DERR_DRIVERINTERNALERROR )
	{
		MessageBox(0, "Internal Driver Error...Exiting", 0, 0);
		PostQuitMessage(0);
		return true;
	}
	// The device is lost but we can reset and restore it.
	else if( hr == D3DERR_DEVICENOTRESET )
	{
		onLostDevice();
		HR(gD3DDevice->Reset(&m_D3DPP));
		onResetDevice();
		return false;
	}
	else
		return false;
}

bool DirectX9App::CheckDeviceCaps()
{
	D3DCAPS9 caps;
	HR(m_D3DDevice->GetDeviceCaps(&caps));

	// Check for vertex shader version 2.0 support.
	if( caps.VertexShaderVersion < D3DVS_VERSION(2, 0) )
		return false;

	// Check for pixel shader version 2.0 support.
	if( caps.PixelShaderVersion < D3DPS_VERSION(2, 0) )
		return false;

	// Check render target support.
	D3DDISPLAYMODE mode;
	HR(m_D3DObject->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &mode));

	// Windowed.
	if(FAILED(m_D3DObject->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, mode.Format, 
		D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_R32F)))
		return false;
	if(FAILED(m_D3DObject->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, mode.Format,
		D3DFMT_R32F, D3DFMT_D24X8)))
		return false;

	// Fullscreen.
	if(FAILED(m_D3DObject->CheckDeviceFormat(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8, 
		D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, D3DFMT_R32F)))
		return false;
	if(FAILED(m_D3DObject->CheckDepthStencilMatch(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8,
		D3DFMT_R32F, D3DFMT_D24X8)))
		return false;
	return true;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	//If DirectX has not be initialized, return the default window proc.
	if(gD3DDevice != 0)
		return DirectX9App::Instance()->MessageProc(msg, wParam, lParam);
	else
		return DefWindowProc(hWnd, msg, wParam, lParam);
}

void DirectX9App::InitMainWindow()
{
	WNDCLASS wc;
	wc.style         = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc   = WndProc;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = m_hAppInst;
	wc.hIcon         = LoadIcon(0, IDI_APPLICATION);
	wc.hCursor       = LoadCursor(0, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName  = 0;
	wc.lpszClassName = "D3DWndClassName";

	if( !RegisterClass(&wc) )
	{
		MessageBox(0, "RegisterClass FAILED", 0, 0);
		PostQuitMessage(0);
	}

	RECT R = {0, 0, BASE_WINDOW_WIDTH, BASE_WINDOW_HEIGHT};
	AdjustWindowRect(&R, WS_OVERLAPPEDWINDOW, false);
	m_hMainWnd = CreateWindow("D3DWndClassName", m_MainWndCaption.c_str(), 
		WS_OVERLAPPEDWINDOW, 100, 100, R.right, R.bottom, 
		0, 0, m_hAppInst, 0); 

	if( !m_hMainWnd )
	{
		MessageBox(0, "CreateWindow FAILED", 0, 0);
		PostQuitMessage(0);
	}

	ShowWindow(m_hMainWnd, SW_SHOW);
	UpdateWindow(m_hMainWnd);
	m_WindowRect = R;
}

void DirectX9App::InitDirect3D()
{
	//Create the IDirect3D9 object.
	m_D3DObject = Direct3DCreate9(D3D_SDK_VERSION);
	if( !m_D3DObject )
	{
		MessageBox(0, "Direct3DCreate9 FAILED", 0, 0);
		PostQuitMessage(0);
	}


	//Verify hardware support for specified formats in windowed and full screen modes.
	D3DDISPLAYMODE mode;
	m_D3DObject->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &mode);
	HR(m_D3DObject->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, mode.Format, mode.Format, true));
	HR(m_D3DObject->CheckDeviceType(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8, D3DFMT_X8R8G8B8, false));

	//Check for requested vertex processing and pure device.
	D3DCAPS9 caps;
	HR(m_D3DObject->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps));

	DWORD devBehaviorFlags = 0;
	if( caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT )
		devBehaviorFlags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		devBehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

	//If pure device and HW T&L supported
	if( caps.DevCaps & D3DDEVCAPS_PUREDEVICE &&
		devBehaviorFlags & D3DCREATE_HARDWARE_VERTEXPROCESSING)
			devBehaviorFlags |= D3DCREATE_PUREDEVICE;

	//Fill out the D3DPRESENT_PARAMETERS structure.
	m_D3DPP.BackBufferWidth            = 0; 
	m_D3DPP.BackBufferHeight           = 0;
	m_D3DPP.BackBufferFormat           = D3DFMT_UNKNOWN;
	m_D3DPP.BackBufferCount            = 1;
	m_D3DPP.MultiSampleType            = D3DMULTISAMPLE_NONE;
	m_D3DPP.MultiSampleQuality         = 0;
	m_D3DPP.SwapEffect                 = D3DSWAPEFFECT_DISCARD; 
	m_D3DPP.hDeviceWindow              = m_hMainWnd;
	m_D3DPP.Windowed                   = true;
	m_D3DPP.EnableAutoDepthStencil     = true; 
	m_D3DPP.AutoDepthStencilFormat     = D3DFMT_D24S8;
	m_D3DPP.Flags                      = 0;
	m_D3DPP.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
	m_D3DPP.PresentationInterval       = D3DPRESENT_INTERVAL_IMMEDIATE;
	
	//Create the device.
	HR(m_D3DObject->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, m_hMainWnd, devBehaviorFlags, &m_D3DPP, &m_D3DDevice)); 

	//Create sprite interface
	HR(D3DXCreateSprite(m_D3DDevice, &m_SpriteInterface));

	//Create the default font
	//Set font for mouse position
	HR(D3DXCreateFont(gD3DDevice, 30, 0, FW_BOLD, 0, false, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, TEXT("Arial"), &m_Font));
}
	
void DirectX9App::Shutdown()
{
	SAFE_RELEASE(m_Font);
	SAFE_RELEASE(m_SpriteInterface);
	SAFE_RELEASE(m_D3DDevice);
	SAFE_RELEASE(m_D3DObject);
}

LRESULT DirectX9App::MessageProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
	// Is the application in a minimized or maximized state?
	static bool minOrMaxed = false;

	RECT clientRect = {0, 0, 0, 0};
	switch( msg )
	{

	//Window activate/deactivate
	case WM_ACTIVATE:
		if( LOWORD(wParam) == WA_INACTIVE )
			m_AppPaused = true;
		else
			m_AppPaused = false;
		return 0;


	//Window Resize
	case WM_SIZE:
		if( m_D3DDevice )
		{
			m_D3DPP.BackBufferWidth  = LOWORD(lParam);
			m_D3DPP.BackBufferHeight = HIWORD(lParam);

			if( wParam == SIZE_MINIMIZED )
			{
				m_AppPaused = true;
				minOrMaxed = true;
			}
			else if( wParam == SIZE_MAXIMIZED )
			{
				m_AppPaused = false;
				minOrMaxed = true;
				onLostDevice();
				HR(m_D3DDevice->Reset(&m_D3DPP));
				onResetDevice();
			}
			
			else if( wParam == SIZE_RESTORED )
			{
				m_AppPaused = false;

				// Are we restoring from a mimimized or maximized state, and in windowed mode?
				if( minOrMaxed && m_D3DPP.Windowed )
				{
					onLostDevice();
					HR(m_D3DDevice->Reset(&m_D3DPP));
					onResetDevice();
				}
				minOrMaxed = false;
			}
		}
		return 0;


	// Reset everything based on the new window dimensions.
	case WM_EXITSIZEMOVE:
		GetClientRect(m_hMainWnd, &clientRect);
		m_D3DPP.BackBufferWidth  = clientRect.right;
		m_D3DPP.BackBufferHeight = clientRect.bottom;
		onLostDevice();
		HR(m_D3DDevice->Reset(&m_D3DPP));
		onResetDevice();

		return 0;

	//Window close
	case WM_CLOSE:
		DestroyWindow(m_hMainWnd);
		return 0;

	// WM_DESTROY is sent when the window is being destroyed.
	case WM_DESTROY:
		Shutdown();
		PostQuitMessage(0);
		return 0;

	case WM_KEYDOWN:
		return 0;
	}
	return DefWindowProc(m_hMainWnd, msg, wParam, lParam);
}

void DirectX9App::LoadXFile(const std::string& filename, ID3DXMesh** meshOut, std::vector<D3DXMATERIAL>& mtrls, std::vector<IDirect3DTexture9*>& texs)
{

	//Load the .x file from file into a system memory mesh.

	ID3DXMesh* meshSys      = 0;
	ID3DXBuffer* adjBuffer  = 0;
	ID3DXBuffer* mtrlBuffer = 0;
	DWORD numMtrls          = 0;

	HR(D3DXLoadMeshFromX(filename.c_str(), D3DXMESH_SYSTEMMEM, m_D3DDevice,
		&adjBuffer,	&mtrlBuffer, 0, &numMtrls, &meshSys));


	//Find out if the mesh already has normal info

	D3DVERTEXELEMENT9 elems[MAX_FVF_DECL_SIZE];
	HR(meshSys->GetDeclaration(elems));
	
	bool hasNormals = false;
	D3DVERTEXELEMENT9 term = D3DDECL_END();
	for(int i = 0; i < MAX_FVF_DECL_SIZE; ++i)
	{
		// Did we reach D3DDECL_END() {0xFF,0,D3DDECLTYPE_UNUSED, 0,0,0}?
		if(elems[i].Stream == 0xff )
			break;

		if( elems[i].Type == D3DDECLTYPE_FLOAT3 &&
			elems[i].Usage == D3DDECLUSAGE_NORMAL &&
			elems[i].UsageIndex == 0 )
		{
			hasNormals = true;
			break;
		}
	}


	//Change vertex format to VertexPNT.

	D3DVERTEXELEMENT9 elements[64];
	UINT numElements = 0;
	VertexPNT::Decl->GetDeclaration(elements, &numElements);

	ID3DXMesh* temp = 0;
	HR(meshSys->CloneMesh(D3DXMESH_SYSTEMMEM, 
		elements, m_D3DDevice, &temp));
	SAFE_RELEASE(meshSys);
	meshSys = temp;


	//If the mesh did not have normals, generate them.

	if( hasNormals == false)
		HR(D3DXComputeNormals(meshSys, 0));


	//Optimize the mesh.

	HR(meshSys->Optimize(D3DXMESH_MANAGED | 
		D3DXMESHOPT_COMPACT | D3DXMESHOPT_ATTRSORT | D3DXMESHOPT_VERTEXCACHE, 
		(DWORD*)adjBuffer->GetBufferPointer(), 0, 0, 0, meshOut));
	SAFE_RELEASE(meshSys); // Done w/ system mesh.
	SAFE_RELEASE(adjBuffer); // Done with buffer.

	//Extract the materials and load the textures.

	if( mtrlBuffer != 0 && numMtrls != 0 )
	{
		D3DXMATERIAL* d3dxmtrls = (D3DXMATERIAL*)mtrlBuffer->GetBufferPointer();

		for(DWORD i = 0; i < numMtrls; ++i)
		{
			// Save the ith material.  Note that the MatD3D property does not have an ambient
			// value set when its loaded, so just set it to the diffuse value.
			D3DXMATERIAL m;
			m.MatD3D.Ambient = d3dxmtrls[i].MatD3D.Diffuse;
			m.MatD3D.Diffuse = d3dxmtrls[i].MatD3D.Diffuse;
			m.MatD3D.Specular = d3dxmtrls[i].MatD3D.Specular;
			m.MatD3D.Emissive = d3dxmtrls[i].MatD3D.Emissive;
			m.MatD3D.Power = d3dxmtrls[i].MatD3D.Power;
			mtrls.push_back( m );

			// Check if the ith material has an associative texture
			if( d3dxmtrls[i].pTextureFilename != 0 )
			{
				// Yes, load the texture for the ith subset
				IDirect3DTexture9* tex = 0;
				char* texFN = d3dxmtrls[i].pTextureFilename;
				HR(D3DXCreateTextureFromFile(m_D3DDevice, texFN, &tex));

				// Save the loaded texture
				texs.push_back( tex );
			}
			else
			{
				// No texture for the ith subset
				texs.push_back( 0 );
			}
		}
	}
	SAFE_RELEASE(mtrlBuffer); // done w/ buffer
}

//Application Entry Point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE prevInstance, PSTR cmdLine, int showCmd)
{
	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
		_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	//Init Input
	gInput->Init(gApplication->GetMainWnd(), hInstance);

    return DirectX9App::Instance()->Run();
}

void DirectX9App::BeginFrame()
{
	// Clear the back buffer, call BeginScene()
	HR(gD3DDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER , m_ClearColor, 1.0f, 0));
	HR(gD3DDevice->BeginScene());

	//Start sprite operations
	HR(gSprite->Begin(D3DXSPRITE_ALPHABLEND | D3DXSPRITE_SORT_DEPTH_FRONTTOBACK));

}
void DirectX9App::EndFrame()
{
	//End sprite operations
	HR(gSprite->End());

	// EndScene, and Present the back buffer to the display buffer
	HR(gD3DDevice->EndScene())
	HR(gD3DDevice->Present(0, 0, 0, 0));
}

IDirect3DVertexDeclaration9* VertexPNT::Decl;